# metaspector/cli.py
# !/usr/bin/env python3

"""
cli.py
~~~~~~~~~~~~~~~

This module provides the command-line interface for the metaspector library.
"""

import argparse
import json
import logging
import sys
import os

from .inspector import MediaInspector
from ._exceptions import MetaspectorError


def check_file_path(path):
    """Custom type function for argparse to validate if a path exists and is a file."""
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(
            f"The path '{path}' does not exist or is not a file."
        )
    return path


def inspect(args):
    """Handles the 'inspect' subcommand."""
    try:
        # Pass the 'section' argument to the inspect method
        inspector = MediaInspector(args.filepath)
        metadata = inspector.inspect(section=args.section)
        print(json.dumps(metadata, indent=2, ensure_ascii=False))
    except (
        MetaspectorError,
        FileNotFoundError,
        ValueError,
    ) as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"An unexpected error occurred: {e}", file=sys.stderr)
        sys.exit(1)


def export(args):
    """Handles the 'export' subcommand for both cover art and metadata."""
    try:
        inspector = MediaInspector(args.filepath)
        destination_path = args.destination
        base_name = os.path.splitext(os.path.basename(args.filepath))[0]

        # --- Logic for exporting the cover ---
        if args.export_type == "cover":
            cover_art = inspector.get_cover_art()
            if not cover_art:
                print("Error: No cover art found in the file.", file=sys.stderr)
                sys.exit(1)

            extension = ".jpg"
            if cover_art.startswith(b"\x89PNG"):
                extension = ".png"
            default_filename = f"{base_name}{extension}"

            if os.path.isdir(destination_path):
                output_path = os.path.join(destination_path, default_filename)
            else:
                output_path = destination_path

            output_dir = os.path.dirname(output_path)
            if output_dir:
                os.makedirs(output_dir, exist_ok=True)

            with open(output_path, "wb") as f:
                f.write(cover_art)
            print(f"Cover art exported successfully to '{output_path}'.")

        # --- Logic for exporting metadata ---
        elif args.export_type == "meta":
            metadata = inspector.inspect()
            default_filename = f"{base_name}.json"

            if os.path.isdir(destination_path):
                output_path = os.path.join(destination_path, default_filename)
            else:
                output_path = destination_path

            output_dir = os.path.dirname(output_path)
            if output_dir:
                os.makedirs(output_dir, exist_ok=True)

            json_data = json.dumps(metadata, indent=2, ensure_ascii=False)
            with open(output_path, "w", encoding="utf-8") as f:
                f.write(json_data)
            print(f"Metadata exported successfully to '{output_path}'.")

    except (MetaspectorError, FileNotFoundError) as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"An unexpected error occurred: {e}", file=sys.stderr)
        sys.exit(1)


def main():
    """Defines the command-line entry point for the tool."""
    # logging.basicConfig(level=logging.DEBUG,
    #                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(
        description="Inspect and extract metadata from media files.",
        epilog="Use 'metaspector <command> --help' for more information on a specific command.",
    )

    subparsers = parser.add_subparsers(
        dest="command", required=True, help="Available commands"
    )

    # --- Parser for the 'inspect' command ---
    inspect_parser = subparsers.add_parser(
        "inspect",
        help="Inspect a media file and print its metadata as JSON.",
        epilog=(
            "Example: metaspector inspect /path/to/my_video.mp4\n"
            "Example: metaspector inspect /path/to/my_audio.flac --section audio"
        ),
    )
    inspect_parser.add_argument(
        "filepath",
        type=check_file_path,
        help="The full path to the media file to inspect.",
    )
    inspect_parser.add_argument(
        "--section",
        choices=["metadata", "audio", "video", "subtitle"],
        help="Optional: Specify a section to output (e.g., 'audio', 'metadata').",
    )
    inspect_parser.set_defaults(func=inspect)

    # --- Parser for the 'export' command ---
    export_parser = subparsers.add_parser(
        "export",
        help="Export data (cover art or metadata) from a media file.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=(
            "Examples:\n"
            "  # Export cover art to a specific file\n"
            "  metaspector export cover song.flac /covers/art.jpg\n\n"
            "  # Export metadata to a directory (filename will be auto-generated)\n"
            "  metaspector export meta song.mp3 /json_files/"
        ),
    )
    export_parser.add_argument(
        "export_type",
        choices=["cover", "meta"],
        help="The type of data to export: 'cover' for the cover art or 'meta' for metadata JSON.",
    )
    export_parser.add_argument(
        "filepath",
        type=check_file_path,
        help="The full path to the media file to export from.",
    )
    export_parser.add_argument(
        "destination",
        help="The destination path. Can be a full file path or a directory.",
    )
    export_parser.set_defaults(func=export)

    args = parser.parse_args()

    if hasattr(args, "func"):
        args.func(args)


if __name__ == "__main__":
    main()



# !/usr/bin/env python3

import struct

from typing import BinaryIO, Dict, Any, Optional, List
from metaspector.format_handlers.base import BaseMediaParser
from .mp4_utils import _read_box_header, _read_uint32, _read_uint64
from .mp4_boxes import MP4BoxParser
from metaspector.matrices.language_matrix import get_long_language_name


class Mp4Parser(BaseMediaParser):
    """
    Parses MP4/M4V files to extract audio, video, and subtitle track metadata.
    """

    class _TrackCharacteristics(MP4BoxParser._TrackCharacteristics):
        """
        Inherits _TrackCharacteristics from MP4BoxParser.
        This local class is kept for type hinting convenience within Mp4Parser.
        """

        pass

    def __init__(self):
        self.audio_tracks: List[Dict[str, Any]] = []
        self.subtitle_tracks: List[Dict[str, Any]] = []
        self.video_tracks: List[Dict[str, Any]] = []
        self.metadata: Dict[str, Any] = {}
        self.moov_duration: Optional[int] = None
        self.moov_timescale: Optional[int] = None

    def parse(self, f: BinaryIO) -> Dict[str, Any]:
        """Parses the MP4 file from the given binary stream."""
        self.audio_tracks = []
        self.subtitle_tracks = []
        self.video_tracks = []
        self.metadata = {}
        self.moov_duration = None
        self.moov_timescale = None

        file_size = f.seek(0, 2)
        f.seek(0)

        current_pos = 0
        while current_pos < file_size:
            f.seek(current_pos)
            box_type, _, _, box_end = _read_box_header(f)
            if not box_type:
                break

            if box_type == b"moov":
                self._parse_moov(f, box_end)
                if (
                    self.moov_duration
                    and self.moov_timescale
                    and self.moov_timescale > 0
                ):
                    duration_in_seconds = self.moov_duration / self.moov_timescale
                    self.metadata["duration_seconds"] = duration_in_seconds
                    for track in self.audio_tracks:
                        track["duration_seconds"] = duration_in_seconds
                    for track in self.video_tracks:
                        track["duration_seconds"] = duration_in_seconds
                break
            elif box_type == b"meta":
                self.metadata.update(MP4BoxParser.parse_meta(f, box_end))

            current_pos = box_end

        if self.metadata.get("media_type") == 1 and "content_rating" in self.metadata:
            self.metadata.pop("content_rating")

        final_audio_tracks = []
        for track in self.audio_tracks:
            ordered_track = self._order_audio_track(track)
            final_audio_tracks.append(ordered_track)

        final_subtitle_tracks = []
        for track in self.subtitle_tracks:
            ordered_track = self._order_subtitle_track(track)
            final_subtitle_tracks.append(ordered_track)

        final_video_tracks = []
        for track in self.video_tracks:
            ordered_track = self._order_video_track(track)
            final_video_tracks.append(ordered_track)

        total_bitrate_kbps = 0
        for track in final_video_tracks:
            total_bitrate_kbps += track.get("bitrate_kbps") or 0
        for track in final_audio_tracks:
            total_bitrate_kbps += track.get("bitrate_kbps") or 0

        if total_bitrate_kbps > 0:
            self.metadata["bitrate_kbps"] = total_bitrate_kbps

        final_metadata = self._process_metadata_for_output(self.metadata)

        return {
            "metadata": final_metadata,
            "video": final_video_tracks,
            "audio": final_audio_tracks,
            "subtitle": final_subtitle_tracks,
        }

    def get_cover_art(self, f: BinaryIO) -> Optional[bytes]:
        """
        Extracts the raw cover art from the MP4 file by finding the 'covr' atom.
        """
        file_size = f.seek(0, 2)
        f.seek(0)

        def search_in_container(
            container_start: int, container_end: int
        ) -> Optional[bytes]:
            current_pos = container_start
            while current_pos < container_end:
                f.seek(current_pos)
                box_type, _, box_start, box_end = _read_box_header(f)
                if not box_type or box_end > container_end:
                    break
                if box_type in (
                    b"moov",
                    b"udta",
                    b"meta",
                    b"ilst",
                    b"\xa9nam",
                    b"name",
                    b"titl",
                ):
                    search_start = box_start + 8
                    if box_type == b"meta":
                        search_start += 4
                    result = search_in_container(search_start, box_end)
                    if result:
                        return result
                if box_type == b"covr":
                    data_pos = box_start + 8
                    while data_pos < box_end:
                        f.seek(data_pos)
                        data_box_type, _, data_box_start, data_box_end = (
                            _read_box_header(f)
                        )
                        if not data_box_type or data_box_end > box_end:
                            break
                        if data_box_type == b"data":
                            f.seek(data_box_start + 8)
                            f.read(8)
                            return f.read(data_box_end - f.tell())
                        data_pos = data_box_end
                current_pos = box_end
            return None

        return search_in_container(0, file_size)

    def _order_audio_track(self, track: Dict[str, Any]) -> Dict[str, Any]:
        """Reorders audio track fields to place dolby_atmos after total_samples."""
        ordered_keys = [
            "index",
            "handler_name",
            "language",
            "internationalized_language",
            "internationalized_language_long",
            "codec",
            "codec_tag_string",
            "channels",
            "channel_layout",
            "sample_rate",
            "bits_per_sample",
            "bitrate_kbps",
            "duration_seconds",
            "total_samples",
            "dolby_atmos",
            "main_program_content",
            "original_content",
            "dubbed_translation",
            "voice_over_translation",
            "language_translation",
            "describes_video_for_accessibility",
            "enhances_speech_intelligibility",
            "auxiliary_content",
        ]

        ordered_dict = {}
        for key in ordered_keys:
            if key in track:
                ordered_dict[key] = track.pop(key)

        ordered_dict.update(track)
        return ordered_dict

    def _order_subtitle_track(self, track: Dict[str, Any]) -> Dict[str, Any]:
        """Reorders subtitle track fields for consistent output."""
        ordered_keys = [
            "index",
            "handler_name",
            "language",
            "internationalized_language",
            "internationalized_language_long",
            "codec",
            "duration_seconds",
            "main_program_content",
            "original_content",
            "auxiliary_content",
            "forced_only",
            "language_translation",
            "easy_to_read",
            "describes_music_and_sound",
            "transcribes_spoken_dialog",
        ]

        ordered_dict = {}
        for key in ordered_keys:
            if key in track:
                ordered_dict[key] = track.pop(key)

        ordered_dict.update(track)
        return ordered_dict

    def _order_video_track(self, track: Dict[str, Any]) -> Dict[str, Any]:
        """Reorders video track fields for consistent output."""
        ordered_keys = [
            "index",
            "handler_name",
            "language",
            "internationalized_language",
            "internationalized_language_long",
            "codec",
            "codec_tag_string",
            "profile",
            "profile_level",
            "width",
            "height",
            "frame_rate",
            "bitrate_kbps",
            "duration_seconds",
            "total_samples",
            "hdr_format",
            "pixel_format",
            "chroma_location",
            "color_space",
            "color_transfer",
            "color_primaries",
            "transfer_characteristics",
            "matrix_coefficients",
            "color_range",
            "dolby_vision",
            "dolby_vision_profile",
            "dolby_vision_level",
            "dolby_vision_sdr_compatible",
            "main_program_content",
            "original_content",
            "auxiliary_content",
        ]

        ordered_dict = {}
        for key in ordered_keys:
            if key in track:
                ordered_dict[key] = track.pop(key)

        ordered_dict.update(track)
        return ordered_dict

    def _process_metadata_for_output(
        self, raw_metadata: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Reorders and processes metadata fields for consistent output.
        """
        processed_metadata = {}
        output_order = [
            "title",
            "artist",
            "album",
            "album_artist",
            "track_number",
            "track_total",
            "disc_number",
            "disc_total",
            "genre",
            "duration_seconds",
            "bitrate_kbps",
            "release_date",
            "publisher",
            "isrc",
            "barcode",
            "upc",
            "media_type",
            "hd_video",
            "hd_video_definition",
            "hd_video_definition_level",
            "itunesadvisory",
            "rating_age_classification",
            "rating_label",
            "rating_system",
            "rating_unit",
            "replaygain_track_gain",
            "replaygain_album_gain",
            "lyrics",
            "composer",
            "copyright",
            "has_cover_art",
            "cover_art_mime",
            "cover_art_dimensions",
            "comment",
            "encoder",
            "performer",
            "language",
            "record_company",
            "description",
            "long_description",
            "sort_name",
            "sort_artist",
            "sort_album",
            "compilation",
            "gapless_playback",
            "content_id",
            "owner",
            "purchase_date",
            "itunes_account",
            "itunes_country",
            "artist_id",
            "playlist_id",
            "geID",
            "composer_id",
            "external_id",
            "itun_compilation",
        ]

        temp_data = {}
        for key in list(raw_metadata.keys()):
            temp_data[key] = raw_metadata.pop(key)

        for key_in_order in output_order:
            if key_in_order in temp_data:
                processed_metadata[key_in_order] = temp_data.pop(key_in_order)

        processed_metadata.update(temp_data)

        if "track_total" in processed_metadata:
            processed_metadata["track_total"] = int(processed_metadata["track_total"])
        if "disc_total" in processed_metadata:
            processed_metadata["disc_total"] = int(processed_metadata["disc_total"])
        if "itunesadvisory" in processed_metadata:
            processed_metadata["itunesadvisory"] = int(
                processed_metadata["itunesadvisory"]
            )

        return processed_metadata

    def _parse_moov(self, f: BinaryIO, moov_end: int):
        """Orchestrates the parsing of all boxes within the 'moov' box."""
        current_pos = f.tell()
        while current_pos < moov_end:
            f.seek(current_pos)
            box_type, _, _, box_end = _read_box_header(f)
            if not box_type or box_end > moov_end:
                break
            if box_type == b"mvhd":
                mvhd_data = MP4BoxParser.parse_mvhd(f, box_end)
                self.moov_duration = mvhd_data.get("duration")
                self.moov_timescale = mvhd_data.get("timescale")
            elif box_type == b"trak":
                self._parse_trak(f, box_end)
            elif box_type == b"udta":
                udta_current_pos = f.tell()
                while udta_current_pos < box_end:
                    f.seek(udta_current_pos)
                    udta_child_type, _, _, udta_child_end = _read_box_header(f)
                    if not udta_child_type or udta_child_end > box_end:
                        break
                    if udta_child_type == b"meta":
                        self.metadata.update(MP4BoxParser.parse_meta(f, udta_child_end))
                        break
                    udta_current_pos = udta_child_end
            elif box_type == b"meta":
                self.metadata.update(MP4BoxParser.parse_meta(f, box_end))
            current_pos = box_end
        f.seek(moov_end)

    def _parse_trak(self, f: BinaryIO, trak_end: int):
        """Parses a 'trak' box and assembles track information."""
        index: Optional[int] = None
        i18n_lang: Optional[str] = None
        hdlr_details: Dict[str, Any] = {}
        mdhd_details: Dict[str, Any] = {}
        track_chars = self._TrackCharacteristics()
        audio_info: Dict[str, Any] = {}
        video_info: Dict[str, Any] = {}
        first_chunk_offset: Optional[int] = None
        first_sample_size: Optional[int] = None
        total_sample_size: int = 0
        descriptive_track_name: Optional[str] = None
        total_samples: Optional[int] = None
        subtitle_codec: Optional[str] = None

        current_pos = f.tell()
        while current_pos < trak_end:
            f.seek(current_pos)
            box_type, _, box_start, box_end = _read_box_header(f)
            if not box_type or box_end > trak_end:
                break

            if box_type == b"tkhd":
                index = MP4BoxParser.parse_tkhd(f, box_end)
                if index is not None:
                    index -= 1
            elif box_type == b"udta":
                udta_chars, udta_name = MP4BoxParser.parse_udta(f, box_end)
                track_chars = udta_chars
                if udta_name and descriptive_track_name is None:
                    descriptive_track_name = udta_name
                f.seek(box_end)
            elif box_type == b"mdia":
                mdia_pos = f.tell()
                while mdia_pos < box_end:
                    f.seek(mdia_pos)
                    mdia_type, _, _, mdia_end = _read_box_header(f)
                    if not mdia_type or mdia_end > box_end:
                        break

                    if mdia_type == b"mdhd":
                        mdhd_details = MP4BoxParser.parse_mdhd(f, mdia_end)
                    elif mdia_type == b"hdlr":
                        hdlr_details = MP4BoxParser.parse_hdlr(f, mdia_end)
                    elif mdia_type == b"elng":
                        i18n_lang = MP4BoxParser.parse_elng(f, mdia_end)
                    elif mdia_type == b"minf":
                        minf_pos = f.tell()
                        while minf_pos < mdia_end:
                            f.seek(minf_pos)
                            minf_type, _, _, minf_end = _read_box_header(f)
                            if not minf_type or minf_end > mdia_end:
                                break
                            if minf_type == b"stbl":
                                stbl_pos = f.tell()
                                while stbl_pos < minf_end:
                                    f.seek(stbl_pos)
                                    stbl_type, _, stbl_box_start, stbl_end = (
                                        _read_box_header(f)
                                    )
                                    if not stbl_type or stbl_end > minf_end:
                                        break

                                    if stbl_type == b"stsd":
                                        handler_type_from_hdlr = hdlr_details.get(
                                            "type"
                                        )
                                        if handler_type_from_hdlr == b"soun":
                                            audio_stsd_info = (
                                                MP4BoxParser.parse_stsd_audio(
                                                    f, stbl_end
                                                )
                                            )
                                            audio_info.update(audio_stsd_info)
                                        elif handler_type_from_hdlr == b"vide":
                                            video_stsd_info = (
                                                MP4BoxParser.parse_stsd_video(
                                                    f, stbl_end
                                                )
                                            )
                                            video_info.update(video_stsd_info)
                                        elif handler_type_from_hdlr in (
                                            b"sbtl",
                                            b"subt",
                                            b"clcp",
                                        ):
                                            (
                                                subtitle_codec,
                                                subtitle_stsd_name,
                                            ) = MP4BoxParser.parse_stsd_subtitle(
                                                f, stbl_end
                                            )
                                            if subtitle_stsd_name:
                                                descriptive_track_name = (
                                                    subtitle_stsd_name
                                                )
                                    elif stbl_type == b"stsz":
                                        f.seek(stbl_box_start + 8 + 4)
                                        uniform_size = _read_uint32(f)
                                        sample_count = _read_uint32(f)
                                        if sample_count is not None:
                                            total_samples = sample_count
                                            if uniform_size != 0:
                                                if first_sample_size is None:
                                                    first_sample_size = uniform_size
                                                total_sample_size = (
                                                    uniform_size * sample_count
                                                )
                                            else:
                                                sizes_data = f.read(sample_count * 4)
                                                if len(sizes_data) == sample_count * 4:
                                                    sizes = struct.unpack(
                                                        f">{sample_count}I", sizes_data
                                                    )
                                                    total_sample_size = sum(sizes)
                                                    if (
                                                        first_sample_size is None
                                                        and sizes
                                                    ):
                                                        first_sample_size = sizes[0]
                                    elif (
                                        stbl_type == b"stco"
                                        and first_chunk_offset is None
                                    ):
                                        f.seek(stbl_box_start + 8 + 4)
                                        entry_count = _read_uint32(f)
                                        if entry_count is not None and entry_count > 0:
                                            first_chunk_offset = _read_uint32(f)
                                    elif (
                                        stbl_type == b"co64"
                                        and first_chunk_offset is None
                                    ):
                                        f.seek(stbl_box_start + 8 + 4)
                                        entry_count = _read_uint32(f)
                                        if entry_count is not None and entry_count > 0:
                                            first_chunk_offset = _read_uint64(f)
                                    f.seek(stbl_end)
                                    stbl_pos = stbl_end
                            f.seek(minf_end)
                            minf_pos = minf_end
                    elif mdia_type == b"meta":
                        meta_content_start_pos = f.tell()
                        f.read(4)
                        temp_meta_pos = meta_content_start_pos + 4
                        while temp_meta_pos < mdia_end:
                            f.seek(temp_meta_pos)
                            (
                                meta_sub_child_type,
                                _,
                                meta_sub_child_start,
                                meta_sub_child_end,
                            ) = _read_box_header(f)
                            if not meta_sub_child_type or meta_sub_child_end > mdia_end:
                                break
                            if meta_sub_child_type == b"ilst":
                                ilst_current_pos = meta_sub_child_start + 8
                                while ilst_current_pos < meta_sub_child_end:
                                    f.seek(ilst_current_pos)
                                    item_type, _, item_start, item_end = (
                                        _read_box_header(f)
                                    )
                                    if not item_type or item_end > meta_sub_child_end:
                                        break
                                    item_child_pos = item_start + 8
                                    while item_child_pos < item_end:
                                        f.seek(item_child_pos)
                                        data_type, _, _, data_end = _read_box_header(f)
                                        if data_type == b"data":
                                            f.read(8)
                                            potential_name = MP4BoxParser.parse_qtss(
                                                f, data_end
                                            )
                                            if (
                                                potential_name
                                                and item_type == b"\xa9nam"
                                            ):
                                                descriptive_track_name = potential_name
                                                break
                                        item_child_pos = data_end
                                    if descriptive_track_name:
                                        break
                                    ilst_current_pos = item_end
                            if descriptive_track_name:
                                break
                            temp_meta_pos = meta_sub_child_end
                    f.seek(mdia_end)
                    mdia_pos = mdia_end
            current_pos = box_end

        lang = mdhd_details.get("lang", "und")
        i18n_lang_long = get_long_language_name(i18n_lang if i18n_lang else lang)

        track_duration = mdhd_details.get("duration")
        track_timescale = mdhd_details.get("timescale")
        handler_type = hdlr_details.get("type")
        final_track_name = (
            descriptive_track_name
            if descriptive_track_name
            else hdlr_details.get("name")
        )

        if handler_type == b"soun":
            if track_duration and track_timescale and total_sample_size > 0:
                duration_in_seconds = track_duration / track_timescale
                if duration_in_seconds > 0:
                    bitrate_bps = (total_sample_size * 8) / duration_in_seconds
                    audio_info["bitrate_kbps"] = int(bitrate_bps / 1000)

            if total_samples is not None:
                audio_info["total_samples"] = total_samples

            if audio_info.get("codec") == "eac3":
                has_atmos = False
                if (
                    first_chunk_offset is not None
                    and first_sample_size is not None
                    and first_sample_size > 0
                ):
                    original_file_pos = f.tell()
                    try:
                        f.seek(first_chunk_offset)
                        frame_data = f.read(first_sample_size)
                        if b"\x0b\x77" in frame_data or b"\x1f\xee" in frame_data:
                            has_atmos = True
                    finally:
                        f.seek(original_file_pos)
                audio_info["dolby_atmos"] = has_atmos

            self.audio_tracks.append(
                {
                    "index": index or 0,
                    "handler_name": final_track_name,
                    "language": lang,
                    "internationalized_language": i18n_lang,
                    "internationalized_language_long": i18n_lang_long,
                    **audio_info,
                    "main_program_content": track_chars.main_program_content,
                    "original_content": track_chars.original_content,
                    "dubbed_translation": track_chars.dubbed_translation,
                    "voice_over_translation": track_chars.voice_over_translation,
                    "language_translation": track_chars.language_translation,
                    "describes_video_for_accessibility": track_chars.describes_video_for_accessibility,
                    "enhances_speech_intelligibility": track_chars.enhances_speech_intelligibility,
                    "auxiliary_content": track_chars.auxiliary_content,
                }
            )


        elif handler_type in (b"sbtl", b"subt", b"clcp"):

            subtitle_track_info = {
                "index": index or 0,
                "handler_name": final_track_name,
                "language": lang,
                "internationalized_language": i18n_lang,
                "internationalized_language_long": i18n_lang_long,
                "codec": subtitle_codec,
                "main_program_content": track_chars.main_program_content,
                "original_content": track_chars.original_content,
                "auxiliary_content": track_chars.auxiliary_content,
                "forced_only": track_chars.forced_only,
                "language_translation": track_chars.language_translation,
                "easy_to_read": track_chars.easy_to_read,
                "describes_music_and_sound": track_chars.describes_music_and_sound,
                "transcribes_spoken_dialog": track_chars.transcribes_spoken_dialog,

            }

            if track_duration and track_timescale and track_timescale > 0:
                subtitle_track_info["duration_seconds"] = (

                        track_duration / track_timescale

                )

            self.subtitle_tracks.append(subtitle_track_info)

        elif handler_type == b"vide":
            if track_duration and track_timescale:
                duration_in_seconds = track_duration / track_timescale
                if duration_in_seconds > 0:
                    if total_sample_size > 0:
                        bitrate_bps = (total_sample_size * 8) / duration_in_seconds
                        video_info["bitrate_kbps"] = int(bitrate_bps / 1000)
                    if total_samples and total_samples > 0:
                        frame_rate = total_samples / duration_in_seconds
                        video_info["frame_rate"] = round(frame_rate, 3)

            if total_samples is not None:
                video_info["total_samples"] = total_samples

            self.video_tracks.append(
                {
                    "index": index or 0,
                    "handler_name": final_track_name,
                    "language": lang,
                    "internationalized_language": i18n_lang,
                    "internationalized_language_long": i18n_lang_long,
                    **video_info,
                    "main_program_content": track_chars.main_program_content,
                    "original_content": track_chars.original_content,
                    "auxiliary_content": track_chars.auxiliary_content,
                }
            )

        f.seek(trak_end)



# metaspector/matrices/rating_matrix.py
# !/usr/bin/env python3

from typing import List, Tuple, Optional


def get_ratings_matrix() -> List[Tuple[str, str, int]]:
    """Returns the comprehensive ratings matrices."""
    ratings_matrix_references = [
        # Australia Ratings
        ("au-movie", "E", 0),
        ("au-movie", "G", 0),
        ("au-movie", "PG", 8),
        ("au-movie", "M", 15),
        ("au-movie", "MA 15+", 15),
        ("au-movie", "R18+", 18),
        ("au-movie", "X18+", 18),
        ("au-movie", "UNRATED", 18),
        ("au-tv", "P", 0),
        ("au-tv", "C", 8),
        ("au-tv", "G", 13),
        ("au-tv", "PG", 8),
        ("au-tv", "M", 15),
        ("au-tv", "MA 15+", 15),
        ("au-tv", "AV 15+", 15),
        ("au-tv", "UNRATED", 18),
        # Canada Ratings
        ("ca-movie", "G", 0),
        ("ca-movie", "PG", 8),
        ("ca-movie", "14", 14),
        ("ca-movie", "14A", 14),
        ("ca-movie", "18", 18),
        ("ca-movie", "18A", 18),
        ("ca-movie", "R", 18),
        ("ca-movie", "E", 18),
        ("ca-movie", "UNRATED", 18),
        ("ca-tv", "C", 0),
        ("ca-tv", "C8", 8),
        ("ca-tv", "G", 13),
        ("ca-tv", "PG", 18),
        ("ca-tv", "14+", 14),
        ("ca-tv", "18+", 18),
        ("ca-tv", "UNRATED", 18),
        # Germany Ratings
        ("de-movie", "ab 0 Jahren", 0),
        ("de-movie", "ab 6 Jahren", 6),
        ("de-movie", "ab 12 Jahren", 12),
        ("de-movie", "ab 16 Jahren", 16),
        ("de-movie", "ab 18 Jahren", 18),
        # Spain Ratings
        ("es-movie", "ICAA A / TP", 0),
        ("es-movie", "ICAA 7", 7),
        ("es-movie", "ICAA 12", 12),
        ("es-movie", "ICAA 16", 16),
        ("es-movie", "ICAA 18", 18),
        ("es-movie", "ICAA SC", 18),
        # France Ratings
        ("fr-movie", "U", 0),
        ("fr-movie", "10", 10),
        ("fr-movie", "12", 12),
        ("fr-movie", "16", 16),
        ("fr-movie", "18", 18),
        ("fr-movie", "E", 18),
        ("fr-tv", "10", 10),
        ("fr-tv", "12", 12),
        ("fr-tv", "16", 16),
        ("fr-tv", "18", 18),
        # Ireland Ratings
        ("ie-movie", "G", 0),
        ("ie-movie", "PG", 8),
        ("ie-movie", "12", 12),
        ("ie-movie", "15", 15),
        ("ie-movie", "16", 16),
        ("ie-movie", "18", 18),
        ("ie-movie", "UNRATED", 18),
        ("ie-tv", "GA", 0),
        ("ie-tv", "Ch", 8),
        ("ie-tv", "YA", 18),
        ("ie-tv", "PS", 18),
        ("ie-tv", "MA", 18),
        ("ie-tv", "UNRATED", 18),
        # Italy Ratings
        ("it-movie", "T", 0),
        ("it-movie", "VM14", 14),
        ("it-movie", "VM18", 18),
        # Japan Ratings
        ("jp-movie", "G", 0),
        ("jp-movie", "PG12", 12),
        ("jp-movie", "R15+", 15),
        ("jp-movie", "R18+", 18),
        # New Zealand Ratings
        ("nz-movie", "E", 0),
        ("nz-movie", "G", 0),
        ("nz-movie", "PG", 8),
        ("nz-movie", "M", 13),
        ("nz-movie", "R13", 13),
        ("nz-movie", "R15", 15),
        ("nz-movie", "R16", 16),
        ("nz-movie", "R18", 18),
        ("nz-movie", "R", 18),
        ("nz-movie", "UNRATED", 18),
        ("nz-tv", "G", 8),
        ("nz-tv", "PGR", 18),
        ("nz-tv", "AO", 18),
        ("nz-tv", "UNRATED", 18),
        # United Kingdom Ratings
        ("uk-movie", "U", 0),
        ("uk-movie", "Uc", 0),
        ("uk-movie", "PG", 8),
        ("uk-movie", "12", 12),
        ("uk-movie", "12A", 12),
        ("uk-movie", "15", 15),
        ("uk-movie", "18", 18),
        ("uk-movie", "R18", 18),
        ("uk-tv", "Caution", 15),
        ("uk-movie", "E", 18),
        ("uk-movie", "UNRATED", 18),
        # United States Ratings
        ("mpaa", "G", 0),
        ("mpaa", "PG", 8),
        ("mpaa", "PG-13", 13),
        ("mpaa", "R", 17),
        ("mpaa", "NC-17", 17),
        ("mpaa", "UNRATED", 18),
        ("mpaa", "NOT RATED", 18),
        ("us-tv", "TV-Y", 0),
        ("us-tv", "TV-Y7", 7),
        ("us-tv", "TV-G", 0),
        ("us-tv", "TV-PG", 10),
        ("us-tv", "TV-14", 14),
        ("us-tv", "TV-MA", 17),
    ]
    return ratings_matrix_references


# Create a mapping for efficient lookups
_ratings_map = {(system, label): age for system, label, age in get_ratings_matrix()}


def get_age_classification(system: str, label: str) -> Optional[int]:
    """
    Looks up the age classification from a rating system and label.
    """
    if not system or not label:
        return None
    return _ratings_map.get((system, label))



# metaspector/format_handlers/mp4/mp4_utils.py
# !/usr/bin/env python3

import struct
import logging

from typing import BinaryIO, Optional, Tuple

logger = logging.getLogger(__name__)

TRANSFER_CHARACTERISTICS_MAP = {
    1: "bt709",
    4: "bt470m",
    5: "bt470bg",
    6: "smpte170m",
    16: "smpte2084",
    18: "arib-std-b67",
    14: "smpte428",
}

COLOR_PRIMARIES_MAP = {
    1: "bt709",
    5: "smpte170m",
    9: "bt2020",
    10: "smpte428",
    11: "smpte428",
    12: "smpte431",
    13: "smpte432",
}
MATRIX_COEFFICIENTS_MAP = {
    0: "gbr",
    1: "bt709",
    5: "bt470bg",
    6: "smpte170m",
    9: "bt2020nc",
    10: "bt2020c",
    14: "bt2020nc",
    15: "bt2020nc",
}

_CHROMA_LOCATION_MAP = {
    0: "left",
    1: "center",
    2: "topleft",
    3: "top",
    4: "bottomleft",
    5: "bottom",
}

_AV1_CHROMA_LOCATION_MAP = {
    0: "unspecified",
    1: "topleft",
    2: "left",
}


def _read_uint8(f: BinaryIO) -> Optional[int]:
    b = f.read(1)
    if not b:
        return None
    return struct.unpack(">B", b)[0]


def _read_uint16(f: BinaryIO) -> Optional[int]:
    b = f.read(2)
    if len(b) < 2:
        return None
    return struct.unpack(">H", b)[0]


def _read_uint32(f: BinaryIO) -> Optional[int]:
    b = f.read(4)
    if len(b) < 4:
        logger.debug(
            f"DEBUG_READ: _read_uint32: EOF or not enough bytes at {f.tell() - len(b)}. Bytes read: {b!r}"
        )
        return None
    return struct.unpack(">I", b)[0]


def _read_uint64(f: BinaryIO) -> Optional[int]:
    b = f.read(8)
    if len(b) < 8:
        logger.debug(
            f"DEBUG_READ: _read_uint64: EOF or not enough bytes at {f.tell() - len(b)}. Bytes read: {b!r}"
        )
        return None
    return struct.unpack(">Q", b)[0]


def _read_box_header(f: BinaryIO) -> Tuple[Optional[bytes], int, int, int]:
    """
    Reads an MP4 box header and returns (type, size, start_pos, end_pos).
    Handles both 32-bit and 64-bit box sizes.
    """
    box_start = f.tell()
    # logger.debug(f"DEBUG_BOX_HEADER: Attempting to read box header at position {box_start}") # Enable this for very verbose logging

    size_32 = _read_uint32(f)
    if size_32 is None:
        logger.debug(
            f"DEBUG_BOX_HEADER: Failed to read 32-bit size at {box_start}. Returning None."
        )
        return None, 0, 0, 0

    box_type_bytes = f.read(4)
    if len(box_type_bytes) < 4:
        logger.debug(
            f"DEBUG_BOX_HEADER: Failed to read 4-byte box type at {f.tell() - len(box_type_bytes)}. Returning None."
        )
        return None, 0, 0, 0

    if size_32 == 1:  # Extended size (64-bit)
        size_64 = _read_uint64(f)
        if size_64 is None:
            logger.debug(
                f"DEBUG_BOX_HEADER: Failed to read 64-bit size for large box at {f.tell() - len(box_type_bytes) - 8}. Returning None."
            )
            return None, 0, 0, 0
        box_size = size_64
    else:  # 32-bit size
        box_size = size_32

    if box_size == 0:  # Box extends to end of file
        file_size = f.seek(0, 2)
        f.seek(box_start)  # Reset position back to start of box before computing end
        box_end = file_size
        logger.debug(
            f"DEBUG_BOX_HEADER: Found box '{box_type_bytes.decode(errors='replace')}' with size 0 (extends to EOF). End: {box_end}"
        )
    elif box_size < 8:  # Minimum box size is 8 bytes (size + type)
        logger.debug(
            f"DEBUG_BOX_HEADER: Invalid box size {box_size} for box '{box_type_bytes.decode(errors='replace')}' at {box_start}. Returning None."
        )
        return None, 0, 0, 0
    else:
        box_end = box_start + box_size

    return box_type_bytes, box_size, box_start, box_end


def _decode_qt_language_code(lang_bits: int) -> str:
    """Decodes a 15-bit QuickTime language code into a 3-letter string."""
    if not (0 <= lang_bits < 32768):
        return "und"
    chars = [(lang_bits >> 10) & 0x1F, (lang_bits >> 5) & 0x1F, lang_bits & 0x1F]
    return "".join(chr(c + 0x60) for c in chars)



# metaspector/format_handlers/flac/flac.py
# !/usr/bin/env python3

import struct
import logging

from typing import BinaryIO, Dict, Any, Optional
from metaspector.format_handlers.base import BaseMediaParser
from .flac_boxes import (
    parse_streaminfo_block,
    parse_vorbis_comment_block,
    parse_picture_block_metadata,
    get_cover_art_data,
)
from .flac_utils import order_audio_track, process_metadata_for_output

logger = logging.getLogger(__name__)


class FlacParser(BaseMediaParser):
    """
    Parses FLAC files to extract audio and metadata.
    Handles standard FLAC metadata blocks using correct block type IDs.
    """

    def __init__(self):
        self.key_map = {
            "title": "title",
            "artist": "artist",
            "album": "album",
            "date": "release_date",
            "genre": "genre",
            "tracknumber": "track_number",
            "discnumber": "disc_number",
            "comment": "comment",
            "composer": "composer",
            "lyrics": "lyrics",
            "performer": "performer",
            "albumartist": "album_artist",
            "description": "description",
            "organization": "record_company",
            "isrc": "isrc",
            "barcode": "barcode",
            "upc": "upc",
            "media": "media_type",
            "encoder": "encoder",
            "language": "language",
            "replaygain_track_gain": "replaygain_track_gain",
            "replaygain_track_peak": "replaygain_track_peak",
            "replaygain_album_gain": "replaygain_album_gain",
            "replaygain_album_peak": "replaygain_album_peak",
            "bpm": "tempo",
            "copyright": "copyright",
            "publisher": "publisher",
            "tracktotal": "track_total",
            "totaltracks": "track_total",
            "disctotal": "disc_total",
            "totaldiscs": "disc_total",
            "length": "duration_seconds",
        }

    def parse(self, f: BinaryIO) -> Dict[str, Any]:
        """Parses the FLAC file from the given binary stream."""
        audio_tracks = []
        metadata = {"has_cover_art": False}
        total_metadata_size = 0

        f.seek(0, 2)
        total_file_size = f.tell()
        f.seek(0)

        if f.read(4) != b"fLaC":
            raise ValueError("Not a valid FLAC file: missing 'fLaC' magic bytes.")
        total_metadata_size += 4

        last_block = False
        while not last_block:
            current_pos = f.tell()
            header_bytes = f.read(4)
            if len(header_bytes) < 4:
                break

            header_val = struct.unpack(">I", header_bytes)[0]
            last_block = (header_val & 0x80000000) != 0
            block_type = (header_val >> 24) & 0x7F
            block_size = header_val & 0x00FFFFFF
            data_end_pos = current_pos + 4 + block_size
            total_metadata_size += 4 + block_size

            if block_type == 0:  # STREAMINFO
                parse_streaminfo_block(f, audio_tracks)
            elif block_type == 4:  # VORBIS_COMMENT
                parse_vorbis_comment_block(f, block_size, metadata, self.key_map)
            elif block_type == 6:  # PICTURE
                parse_picture_block_metadata(f, metadata)

            f.seek(data_end_pos)

        if audio_tracks:
            if audio_tracks[0].get("duration_seconds"):
                metadata["duration_seconds"] = audio_tracks[0]["duration_seconds"]

            if audio_tracks[0].get("duration_seconds", 0) > 0:
                audio_data_size = (
                    total_file_size - total_metadata_size
                    if total_file_size > total_metadata_size
                    else 0
                )
                if audio_data_size > 0:
                    bitrate = (audio_data_size * 8) / audio_tracks[0][
                        "duration_seconds"
                    ]
                    bitrate_kbps = int(bitrate / 1000)
                    audio_tracks[0]["bitrate_kbps"] = bitrate_kbps
                    metadata["bitrate_kbps"] = bitrate_kbps

        return {
            "metadata": process_metadata_for_output(metadata),
            "audio": [
                order_audio_track(track, i) for i, track in enumerate(audio_tracks)
            ],
            "video": [],
            "subtitle": [],
        }

    def get_cover_art(self, f: BinaryIO) -> Optional[bytes]:
        """
        Efficiently finds and extracts the raw cover art data from a FLAC file
        using a robust, stream-based parsing strategy.
        """
        f.seek(0)
        return get_cover_art_data(f)



# metaspector/format_handlers/base.py
# !/usr/bin/env python3

from abc import ABC, abstractmethod
from typing import BinaryIO, Dict, List, Any


class BaseMediaParser(ABC):
    """
    Abstract base class for media file format parsers.
    Defines the interface that all concrete media parsers must implement.
    """

    @abstractmethod
    def parse(self, f: BinaryIO) -> Dict[str, List]:
        """
        Parses the media file from the given binary stream and
        returns a dictionary of extracted track metadata.

        Args:
            f: A binary file-like object (e.g., an open file handle)
               positioned at the beginning of the relevant data.

        Returns:
            A dictionary with 'audio' and 'subtitle' keys, each containing
            a list of dictionaries, where each inner dictionary represents
            metadata for a track.
            Example: {"audio": [...], "subtitle": [...]}
        """
        pass



# metaspector/format_handlers/mp3/mp3_utils.py
# !/usr/bin/env python3

import struct
import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)

MPEG_VERSIONS = {0b11: "MPEG-1", 0b10: "MPEG-2", 0b00: "MPEG-2.5"}

LAYER_DESCRIPTIONS = {0b11: "Layer I", 0b10: "Layer II", 0b01: "Layer III"}

BITRATE_TABLES = {
    "MPEG-1_Layer-I": {
        1: 32,
        2: 64,
        3: 96,
        4: 128,
        5: 160,
        6: 192,
        7: 224,
        8: 256,
        9: 288,
        10: 320,
        11: 352,
        12: 384,
        13: 416,
        14: 448,
    },
    "MPEG-1_Layer-II": {
        1: 32,
        2: 48,
        3: 56,
        4: 64,
        5: 80,
        6: 96,
        7: 112,
        8: 128,
        9: 160,
        10: 192,
        11: 224,
        12: 256,
        13: 320,
        14: 384,
    },
    "MPEG-1_Layer-III": {
        1: 32,
        2: 40,
        3: 48,
        4: 56,
        5: 64,
        6: 80,
        7: 96,
        8: 112,
        9: 128,
        10: 160,
        11: 192,
        12: 224,
        13: 256,
        14: 320,
    },
    "MPEG-2_Layer-I": {
        1: 32,
        2: 48,
        3: 56,
        4: 64,
        5: 80,
        6: 96,
        7: 112,
        8: 128,
        9: 144,
        10: 160,
        11: 176,
        12: 192,
        13: 224,
        14: 256,
    },
    "MPEG-2_Layer-II": {
        1: 8,
        2: 16,
        3: 24,
        4: 32,
        5: 40,
        6: 48,
        7: 56,
        8: 64,
        9: 80,
        10: 96,
        11: 112,
        12: 128,
        13: 144,
        14: 160,
    },
    "MPEG-2_Layer-III": {
        1: 8,
        2: 16,
        3: 24,
        4: 32,
        5: 40,
        6: 48,
        7: 56,
        8: 64,
        9: 80,
        10: 96,
        11: 112,
        12: 128,
        13: 144,
        14: 160,
    },
    "MPEG-2.5_Layer-I": {
        1: 32,
        2: 48,
        3: 56,
        4: 64,
        5: 80,
        6: 96,
        7: 112,
        8: 128,
        9: 144,
        10: 160,
        11: 176,
        12: 192,
        13: 224,
        14: 256,
    },
    "MPEG-2.5_Layer-II": {
        1: 8,
        2: 16,
        3: 24,
        4: 32,
        5: 40,
        6: 48,
        7: 56,
        8: 64,
        9: 80,
        10: 96,
        11: 112,
        12: 128,
        13: 144,
        14: 160,
    },
    "MPEG-2.5_Layer-III": {
        1: 8,
        2: 16,
        3: 24,
        4: 32,
        5: 40,
        6: 48,
        7: 56,
        8: 64,
        9: 80,
        10: 96,
        11: 112,
        12: 128,
        13: 144,
        14: 160,
    },
}

SAMPLE_RATE_TABLES = {
    "MPEG-1": {0b00: 44100, 0b01: 48000, 0b10: 32000},
    "MPEG-2": {0b00: 22050, 0b01: 24000, 0b10: 16000},
    "MPEG-2.5": {0b00: 11025, 0b01: 12000, 0b10: 8000},
}

CHANNEL_MODES = {
    0b00: "Stereo",
    0b01: "Joint Stereo",
    0b10: "Dual Channel",
    0b11: "Mono",
}

SAMPLES_PER_FRAME = {
    "MPEG-1_Layer-I": 384,
    "MPEG-1_Layer-II": 1152,
    "MPEG-1_Layer-III": 1152,
    "MPEG-2_Layer-I": 384,
    "MPEG-2_Layer-II": 1152,
    "MPEG-2_Layer-III": 576,
    "MPEG-2.5_Layer-I": 384,
    "MPEG-2.5_Layer-II": 1152,
    "MPEG-2.5_Layer-III": 576,
}

ID3_KEY_MAP = {
    "TIT2": "title",
    "TPE1": "artist",
    "TALB": "album",
    "TCOM": "composer",
    "TCON": "genre",
    "TDRC": "release_date",
    "TYER": "release_date",
    "TRCK": "track_number",
    "TPOS": "disc_number",
    "TPE2": "album_artist",
    "USLT": "lyrics",
    "TENC": "encoder",
    "COMM": "comment",
    "TCOP": "copyright",
    "TPUB": "publisher",
    "TSOP": "performer",
    "TMPO": "tempo",
    "TSRC": "isrc",
    "UFID": "unique_file_identifier",
    "TSSE": "encoder",
    "TXXX:REPLAYGAIN_TRACK_GAIN": "replaygain_track_gain",
    "TXXX:REPLAYGAIN_TRACK_PEAK": "replaygain_track_peak",
    "TXXX:REPLAYGAIN_ALBUM_GAIN": "replaygain_album_gain",
    "TXXX:REPLAYGAIN_ALBUM_PEAK": "replaygain_album_peak",
    "TXXX:CUSTOM_TRACKTOTAL": "track_total",
    "TXXX:CUSTOM_DISCTOTAL": "disc_total",
    "TXXX:CUSTOM_LENGTH": "duration_seconds",
    "TXXX:CUSTOM_BPM": "tempo",
    "TXXX:CUSTOM_ISRC": "isrc",
    "TXXX:CUSTOM_BARCODE": "barcode",
    "TXXX:CUSTOM_ITUNESADVISORY": "itunesadvisory",
    "TXXX:ORGANIZATION": "record_company",
    "TXXX:UPC": "upc",
    "TXXX:MEDIA": "media_type",
    "TXXX:DESCRIPTION": "description",
    "TXXX:COMMENT": "comment",
    "TXXX:DATE": "release_date",
    "TXXX:PERFORMER": "performer",
    "TLEN": "duration_seconds",
}


def decode_id3_string(data: bytes, encoding_byte: int) -> str:
    """
    Decodes an ID3v2 string based on the specified encoding byte.
    """
    try:
        if encoding_byte == 0x00:
            decoded_string = data.decode("latin-1", errors="replace")
        elif encoding_byte == 0x01:
            decoded_string = data.decode("utf-16", errors="replace")
        elif encoding_byte == 0x02:
            decoded_string = data.decode("utf-16-be", errors="replace")
        elif encoding_byte == 0x03:
            decoded_string = data.decode("utf-8", errors="replace")
        else:
            logger.warning(
                f"Unknown ID3v2 string encoding: {encoding_byte}. Attempting UTF-8."
            )
            decoded_string = data.decode("utf-8", errors="replace")
    except UnicodeDecodeError:
        logger.error(
            f"Failed to decode string with encoding {encoding_byte}. Falling back to latin-1."
        )
        decoded_string = data.decode("latin-1", errors="replace")
    except Exception as e:
        logger.error(
            f"Error during string decoding (encoding {encoding_byte}): {e}. Data: {data!r}"
        )
        return ""
    return decoded_string.replace("\x00", "").strip()


def parse_image_and_update_metadata(
    image_data: bytes, mime_type: str
) -> Dict[str, Any]:
    """
    Parses image data to get dimensions and returns a metadata dictionary.
    """
    metadata = {
        "has_cover_art": True,
        "cover_art_mime": mime_type,
        "cover_art_dimensions": None,
    }

    width, height = None, None
    if mime_type == "image/jpeg":
        sof_markers = [b"\xff\xc0", b"\xff\xc1", b"\xff\xc2"]
        for marker in sof_markers:
            sof_marker_pos = image_data.find(marker)
            if sof_marker_pos != -1 and len(image_data) >= sof_marker_pos + 9:
                try:
                    height = struct.unpack(
                        ">H", image_data[sof_marker_pos + 5 : sof_marker_pos + 7]
                    )[0]
                    width = struct.unpack(
                        ">H", image_data[sof_marker_pos + 7 : sof_marker_pos + 9]
                    )[0]
                    break
                except Exception:
                    pass
    elif mime_type == "image/png":
        if len(image_data) >= 24 and image_data.startswith(b"\x89PNG\r\n\x1a\n"):
            try:
                width = struct.unpack(">I", image_data[16:20])[0]
                height = struct.unpack(">I", image_data[20:24])[0]
            except Exception:
                pass

    if width is not None and height is not None:
        metadata["cover_art_dimensions"] = f"{width}x{height}"

    return metadata
